from .string import str_to_ascii


class Log:
    """
    This class override logging system to fix encoding problem.

    :param string logger_name: The logger name, define in settings.
    """
    logger = None

    def __init__(self, logger_name):
        import logging
        self.logger_name = logger_name
        self.logger = logging.getLogger(logger_name)

    def debug(self, msg):
        """
        Add a log message level 'debug'.

        :param string msg: Message to log
        """
        try:
            self.logger.debug(f"{msg}")
        except:
            self.logger.warning("Error logging format!")
            self.logger.debug(str_to_ascii(f"{msg}"))

    def info(self, msg):
        """
        Add a log message level 'info'.

        :param string msg: Message to log
        """
        try:
            self.logger.info(f"{msg}")
        except:
            self.logger.warning("Error logging format!")
            self.logger.info(str_to_ascii(f"{msg}"))

    def warning(self, msg):
        """
        Add a log message level 'warning'.

        :param string msg: Message to log
        """
        try:
            self.logger.warning(f"{msg}")
        except:
            self.logger.warning("Error logging format!")
            self.logger.warning(str_to_ascii(f"{msg}"))

    def error(self, msg):
        """
        Add a log message level 'error'.

        :param string msg: Message to log
        """
        try:
            self.logger.error(f"{msg}")
        except:
            self.logger.warning("Error logging format!")
            self.logger.error(str_to_ascii(f"{msg}"))

    def critical(self, msg):
        """
        Add a log message level 'critical'.

        :param string msg: Message to log
        """
        try:
            self.logger.critical(f"{msg}")
        except:
            self.logger.warning("Error logging format!")
            self.logger.critical(str_to_ascii(f"{msg}"))


def getLogger():
    """
    This function generate a Log object
    """
    return Log("smart-prospective-api")



import functools
from .log import getLogger
from .exceptions import APIError


def network_try_except(function):
    """
        Wrap a function with a try/except and log on error
    """
    @functools.wraps(function)
    def wrap(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except APIError as e:
            getLogger().error(f"Error from API call: {e}")
            raise e
        except Exception as e:
            getLogger().error(f"Error from API call: {e}")
            raise APIError(f"Error in the call for {function.__name__}: {e}")
    return wrap


def is_login(function):
    """
        Check if the instance has a token (=is_login), if not, attempt to login.
    """
    @functools.wraps(function)
    def wrap(*args, **kwargs):
        if not args[0]:
            getLogger().critical(f"The decorator @is_login must get SPApi instance to be executed")
            raise APIError("Invalid call for @is_login")
        try:
            if not args[0].token:
                args[0].login()
        except Exception as e:
            getLogger().error(f"Failure to login the user")
            raise APIError(f"The user must be logged")
        return function(*args, **kwargs)
    return wrap


def supported_parameters(parameters):
    """
        Check if the given parameters are all supported

        :param parameters [str]: List of str, which represents the supported parameters
    """
    def _supported_parameters(function):
        def wrap(*args, **kwargs):
            invalid_keys = False
            for key, value in kwargs.items():
                if key not in parameters:
                    invalid_keys = True
                    getLogger().error(f"Invalid parameter {key}, not supported for {function.__name__}")
            if invalid_keys:
                raise APIError(f"Cancel call for {function.__name__}")
            return function(*args, **kwargs)
        return wrap
    return _supported_parameters



from requests import get, post
import json
import os
import re

from .exceptions import APIError
from .decorators import network_try_except
from .log import getLogger


def get_server(subpath):
    """
        Return the complete api server url with subpath, by checking environnement variable 'SMART_PROSPECTIVE_SERVER' or use a default one

        :param subpath str: The subpath of the url to use (no need to include '/api' e.g: 'users/add')
        :rtype: str
        :return: The complete url
    """
    if "SMART_PROSPECTIVE_SERVER" in os.environ:
        server_url = os.environ["SMART_PROSPECTIVE_SERVER"]
    else:
        server_url = "https://app.smartprospective.com"
    # Add the /api to url (to save code)
    return f"{server_url}/api{subpath}"


def treat_response(response):
    """
        Treat a request.Response from a get() or a post() by parsing in json the response and checking the status code of the request.

        :param response request.Response: The response to treat
        :rtype: {...}/None
        :raise APIError: If status code is 4XX or 5XX
        :return: The JSON response. None otherwise (invalid JSON)
    """
    try:
        json_response = json.loads(response.text)
    except Exception as err:
        getLogger().error(f"SPApi.network: Invalid json response: {response.text}")
        json_response = None
    if int(response.status_code / 100) in [4, 5]:
        # Error
        getLogger().error(f"SPApi.network: Error from request:\nUrl:{response.url}\nResponse:{json_response}")
        raise APIError(f"Invalid status code response: {response.status_code}")
    return json_response


def treat_file_response(response, filename=None):
    """
        Treat a request.Response with a file given from a get() or a post() by using response.content

        :param response request.Response: The response to treat
        :param filename str/None: The filename (no extension) to create, can include the path (absolute or relative) (will be created if doesn't exist)
        :rtype: str/None
        :raise APIError: If status code is 4XX or 5XX / Folder cannot be created
        :return: The filepath created. None otherwise (invalid response)
    """
    if int(response.status_code / 100) in [4, 5]:
        # Error
        getLogger().error(f"SPApi.network: Error from request:\nUrl:{response.url}\nResponse Status:{response.text}")
        raise APIError(f"Invalid status code response: {response.status_code}")
    try:
        # Complete or Set filename
        try:
            request_filename = re.findall("filename=(.+)", response.headers['content-disposition'])[0]
            if not filename:
                # If filename not given, use the one from the request
                filename = request_filename
            else:
                # If filename given, only contact file extension
                filename += "." + request_filename.split(".")[-1]
        except Exception as e:
            getLogger().warning(f"SPApi.network: Cannot find the filename in the request: {response.headers}, use default filename 'unknown'")
            filename = "unknown"
        # Check folder from filename (create if if needed)
        folder = os.path.dirname(filename)
        if folder and not os.path.isdir(folder):
            os.makedirs(folder)
        with open(filename, "wb") as f:
            # Write the file from the request content
            f.write(response.content)
    except Exception as err:
        getLogger().error(f"SPApi.network: Invalid file response: {err}")
        filename = None
    return filename


@network_try_except
def default_get(token, resource):
    """
        Perform a GET request, insert the token as GET parameter and finally treat the response (JSON)

        :param token str: The token (set in GET parameters)
        :param resource str: The resource to build the url (https://server_url/api/{resource})
        :rtype: {...}/None
        :raise APIError: See treat_response()
        :return: See treat_response()
    """
    url = get_server(f"/{resource}") + f"?token={token}"
    return treat_response(get(url))


@network_try_except
def default_post(parameters, resource, files=None):
    """
        Perform a POST request and finally treat the response (JSON)

        :param parameters {...}: The data to use in the POST request
        :param resource str: The resource to build the url (https://server_url/api/{resource})
        :param files {...}: The files to insert
        :rtype: {...}/None
        :raise APIError: See treat_response()
        :return: See treat_response()
    """
    return treat_response(post(get_server(f"/{resource}"), data=parameters, files=files))


@network_try_except
def post_to_download(parameters, resource, files=None, filename=None):
    """
        Perform a POST request and finally treat the file response (filepath)

        :param parameters {...}: The data to use in the POST request
        :param resource str: The resource to build the url (https://server_url/api/{resource})
        :param files {...}: The files to insert
        :param filename str/None: The filename to create (See treat_file_response())
        :rtype: str/None
        :raise APIError: See treat_file_response()
        :return: See treat_file_response()
    """
    return treat_file_response(post(get_server(f"/{resource}"), data=parameters, files=files), filename=filename)



from django.http import JsonResponse


class APIError(RuntimeError):
    """
    Exception for a API error, can be converted into an error

    :param string msg: The msg
    """
    status = 400

    def __init__(self, msg):
        super(APIError, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        """
        Format for str

        :return: The exception message
        :rtype: string
        """
        return f"APIError: {self.msg}"



def doc():
    print("""This python library is made to connect to Smart Prospective API, in order to manage your:
- Account
- Materials
- Buildings
- Material Groups
- Users
- Medias
- Tags
- Rights
- Right Groups
- PostAccount
- Notifications
- Bills
- Deliveries
- Interventions

Documentation & Samples: https://github.com/smart-prospective/smart-prospective-api
""")


if __name__ == "__main__":
    doc()



import pathlib
from setuptools import setup

# The text of the README file
README = (pathlib.Path(__file__).parent / "README.md").read_text()

setup(name='smart-prospective-api',
      version='1.0.1',
      description='The Official Python client for Smart Prospective API',
      long_description=README,
      long_description_content_type="text/markdown",
      url='https://github.com/smart-prospective/smart-prospective-api',
      author='Smart Prospective',
      author_email='dev@smartprospective.com',
      license='MIT',
      packages=['smart_prospective_api'],
      install_requires=['requests'],
      zip_safe=False)



from .network import *
from .log import getLogger
from .decorators import is_login, supported_parameters


class SPApi():
    token = None
    api_public_key = None
    api_secret_key = None

    def __init__(self, api_public_key, api_secret_key):
        self.api_public_key = api_public_key
        self.api_secret_key = api_secret_key

    def login(self):
        """
            Login the user, in order to get a token using the api_public_key & api_secret_key.
            If the token is already present, no login is performed.

            :raise APIError: On any error
            :rtype: str
            :return: The API token, valid until logout or timeout
        """
        if not self.token:
            self.token = default_post({
                "api_public_key": self.api_public_key,
                "api_secret_key": self.api_secret_key
            }, "login")["token"]
        else:
            getLogger().debug("SPApi: Already connected")
        return self.token

    @is_login
    def logout(self):
        """
            Logout the user, in order to get release the token.
            The token must be set already present, no login is performed.

            :raise APIError: On any error
        """
        default_post({"token": self.token}, "logout")
        self.token = None

    # Users
    @is_login
    def get_users(self):
        """
            Get all the users link to this user.

            :raise APIError: On any error
            :rtype: [{...}]
            :return: The list of users
        """
        return default_get(self.token, "users")["users"]

    @is_login
    @supported_parameters(["first_name", "last_name", "email", "phone", "position", "password", "right_groups", "rights", "picture", "picture_link",
                           "buildings", "materialgroups", "materials"])
    def add_user(self, **kwargs):
        kwargs["token"] = self.token
        user = default_post(kwargs, "users/add")["user"]
        getLogger().info("User {user['name']} created!")
        return user

    @is_login
    def delete_user(self, code):
        default_post({
            "token": self.token,
            "user_code": code
        }, "users/remove")
        getLogger().info("User {code} deleted!")

    # Materials
    @is_login
    def get_materials(self):
        """
            Get all the materials.

            :raise APIError: On any error
            :rtype: [{...}]
            :return: The list of materials
        """
        return default_get(self.token, "materials")["materials"]

    @is_login
    @supported_parameters(["name", "building", "link_code", "picture", "picture_link", "default_media", "time_to_invalid",
                           "users", "division", "rotation", "bg_color", "width", "height", "player_config"])
    def add_material(self, **kwargs):
        kwargs["token"] = self.token
        material = default_post(kwargs, "materials/add")["material"]
        getLogger().info("Material {material['name']} created!")
        return material

    @is_login
    def reboot_material(self, code):
        status = default_post({
            "token": self.token,
            "material_code": code
        }, "materials/reboot")["status"]
        if status:
            getLogger().info("Material {code} will reboot!")
        else:
            getLogger().warning("Material {code} has already been asked to reboot, need to wait!")
        return status

    @is_login
    def refresh_material(self, code):
        status = default_post({
            "token": self.token,
            "material_code": code
        }, "materials/refresh")["status"]
        if status:
            getLogger().info("Material {code} will be refreshed!")
        else:
            getLogger().warning("Material {code} has already been asked to refresh, need to wait!")
        return status

    # Material Groups
    @is_login
    def get_materialgroups(self):
        """
            Get all the materialgroups.

            :raise APIError: On any error
            :rtype: [{...}]
            :return: The list of materialgroups
        """
        return default_get(self.token, "material-groups")["materialgroups"]

    @is_login
    @supported_parameters(["name", "comment", "materials", "buildings", "picture", "picture_link"])
    def add_materialgroup(self, **kwargs):
        kwargs["token"] = self.token
        materialgroup = default_post(kwargs, "material-groups/add")["materialgroup"]
        getLogger().info("Material Group {materialgroup['name']} created!")
        return materialgroup

    @is_login
    def delete_materialgroup(self, code):
        default_post({
            "token": self.token,
            "materialgroup_code": code
        }, "material-groups/remove")
        getLogger().info("User {code} deleted!")

    # Buildings
    @is_login
    def get_buildings(self):
        """
            Get all the buildings.

            :raise APIError: On any error
            :rtype: [{...}]
            :return: The list of buildings
        """
        return default_get(self.token, "buildings")["buildings"]

    # Media

    @is_login
    def get_medias(self):
        """
            Get all the medias.

            :raise APIError: On any error
            :rtype: [{...}]
            :return: The list of medias
        """
        return default_get(self.token, "medias")["medias"]

    @is_login
    @supported_parameters(["category", "name", "comment", "tags", "force_fullscreen", "lock", "buildings", "material_groups", "materials",
                           "format_details", "condition_start", "condition_end", "condition_weather", "condition_time", "frequency",
                           "print_min", "print_max", "animation_start", "animation_end", "male", "female", "both", "age_zero_fifteen",
                           "age_sixteen_twenty_eight", "age_twenty_nine_thirty_six", "age_thirty_seven_fifty", "age_fifty_one_ninty_nine",
                           "interests", "pcs_farmer", "pcs_worker", "pcs_retirees", "pcs_intermediate_professions", "pcs_employee",
                           "pcs_student", "pcs_unemployed", "pcs_craftsmen", "pcs_managment_nd_profession", "vistor_type_new",
                           "vistor_type_frequent", "vistor_type_occasional", "vistor_type_everybody", "specific_duration", "keep_audio",
                           "url", "file", "post_accounts", "webview_details", "webviewtemplate"])
    def add_media(self, category, **kwargs):
        # If file is given -> Double request: 1st Media File Upload -> 2nd add media
        if "file" in kwargs:
            file_input = None
            with open(kwargs["file"], "rb") as f:
                file_input = default_post({
                    "token": self.token
                }, "medias/upload", files={"file": f})["file"]["code"]
            if not file_input:
                getLogger().error(f"SPApi.add_media: Error from file upload\nFile after upload:{file_input}")
                raise APIError(f"Failure on file upload")
            # Field convertion "file" -> "file_input"
            kwargs["file_input"] = file_input
            del kwargs["file"]
        # Field convertion "post_accounts" -> "post_accounts_text"
        if "post_accounts" in kwargs:
            kwargs["post_accounts_text"] = ",".join(kwargs["post_accounts"])
            del kwargs["post_accounts"]
        # Field convertion "tags" -> "tags_text"
        if "tags" in kwargs:
            kwargs["tags_text"] = ",".join(kwargs["tags"])
            del kwargs["tags"]
        # Field convertion "interests" -> "interests_text"
        if "interests" in kwargs:
            kwargs["interests_text"] = ",".join(kwargs["interests"])
            del kwargs["interests"]
        kwargs["token"] = self.token
        media = default_post(kwargs, f"medias/add/{category}")["media"]
        getLogger().info("Media {media['name']} created!")
        return media

    @is_login
    def upload_media_template_file(self, file):
        with open(file, "rb") as f:
            return default_post({
                "token": self.token
            }, "medias/upload/template", files={"file": f})["file"]
        raise APIError(f"Failure on template file upload, maybe the file is not found or not readable")

    @is_login
    def download_final_media(self, media_code, material_code, filename=None):
        return post_to_download({
            "token": self.token,
            "media_code": media_code,
            "material_code": material_code,
        }, "medias/download/final", filename=filename)

    def download_converted_media(self, media_code, filename=None):
        return post_to_download({
            "token": self.token,
            "media_code": media_code
        }, "medias/download/convert", filename=filename)

    def download_src_media(self, media_code, filename=None):
        return post_to_download({
            "token": self.token,
            "media_code": media_code
        }, "medias/download/src", filename=filename)

    @is_login
    def disable_media(self, media_code):
        status = default_post({
            "token": self.token,
            "media_code": media_code
        }, "medias/disable")["status"]
        if status:
            getLogger().info("Media {code} has been disable!")
        else:
            getLogger().warning("Media {code} cannot be disable!")
        return status

    @is_login
    def enable_media(self, media_code):
        status = default_post({
            "token": self.token,
            "media_code": media_code
        }, "medias/enable")["status"]
        if status:
            getLogger().info("Media {code} has been enable!")
        else:
            getLogger().warning("Media {code} cannot be enable!")
        return status

    @is_login
    def delete_media(self, code):
        default_post({
            "token": self.token,
            "media_code": code
        }, "medias/remove")
        getLogger().info("Media {code} deleted!")

    # WebviewTemplate

    @is_login
    def get_webviewtemplates(self):
        """
            Get all the webviewtemplates.

            :raise APIError: On any error
            :rtype: [{...}]
            :return: The list of webviewtemplates
        """
        return default_get(self.token, "webviewtemplates")["webviewtemplates"]
